from __future__ import absolute_import
from __future__ import print_function
try:
    import numpy as np
    import reflex
    from pipeline_product import PipelineProduct
    import pipeline_display
    import re
    import_success = True

except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")

#from copy import deepcopy as dc
import matplotlib.gridspec as gridspec
from matplotlib.patches import Rectangle

class DataPlotterManager:

    recipe_name = "espda_create_linelist"

    """ General methods """

    def addSubplots(self, figure):
        self.gs = gridspec.GridSpec(42, 6)
        self.figure = figure
        self.p = Plot()
        self.p.sub_b = self.p.sub(self.figure, self.gs[34:39,0:1], xlabel=None,
                                  ylabel=None)
        self.p.sub_s = self.p.sub(self.figure, self.gs[40:,:], xlabel=None,
                                  ylabel=None)


    def plotProductsGraphics(self):
        self.vals = []
        self.error_d = False
        self.plotSpec()


    def plotSpec(self):
        self.plotSpecMain()
        self.plotSpecDetail()


    def readFitsData(self, fitsFiles):
        frames = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            frames[frame.category] = PipelineProduct(frame)

        wavel = frames["RSPEC"].readTableColumn(1, "WAVEL")
        self.wavel_min = np.min(wavel)
        self.wavel_max = np.max(wavel)
        self.wavel_mean = np.mean(wavel)

        self.rspec_m = Data(frames["RSPEC"])
        #self.rspec_d = dc(self.rspec_m)
        self.rspec_d = Data(frames["RSPEC"])
        self.rspec_d.select("WAVEL", (self.wavel_mean-10, self.wavel_mean+10))
        try:
            self.rspec_smooth_m = Data(frames["RSPEC_SMOOTH"])
            self.fline_pre_m = Data(frames["FLINE_PRE"], y="PEAK")
            #self.rspec_smooth_d = dc(self.rspec_smooth_m)
            #self.fline_pre_d = dc(self.fline_pre_m)
            self.rspec_smooth_d = Data(frames["RSPEC_SMOOTH"])
            self.fline_pre_d = Data(frames["FLINE_PRE"], y="PEAK")
            self.rspec_smooth_d.select("WAVEL",
                                       (self.wavel_mean-10, self.wavel_mean+10))
            self.fline_pre_d.select("WAVEL",
                                    (self.wavel_mean-10, self.wavel_mean+10))
            self.fline_num = len(self.fline_pre_m.x)
        except:
            pass

    def setInteractiveParameters(self):
        paramList      = list()
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName = "vel-step-min", group = "Smoothing", description =
            "Minimum velocity step to undersample the spectra, corresponding "
            "to the minimum FWHM of lines to be detected [km/s]: [0.3, 1.0e7]"))
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName = "vel-step-max", group = "Smoothing", description =
            "Maximum velocity step to undersample the spectra, corresponding "
            "to the maximum FWHM of lines to be detected [km/s]: [0.3, 1.0e7]"))
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName = "accept",   group = "Line detection", description =
            "Distance between a flux minimum and the adjacent maxima for the "
            "minimum to be accepted as an absorption line, as a multiple of "
            "the flux error in the minimum: [3.0, 1.0e2]"))
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName = "add",      group = "Line detection", description =
            "Wavelengths of additional, manually-selected lines [nm]. "
            "Wavelengths must be separated by commas, with a trailing comma "
            "after the last wavelength."))
        return paramList


    def setWindowTitle(self):
        title = 'ESPRESSO Data Analysis Software: Line Detection'
        return title


    """ Specific methods """

    def addLine(self, point):
        val = self.rspec_d.get_xy(point.xdata, point.ydata)
        if val in self.vals:
            self.vals.remove(val)
        else:
            self.vals.append(val)

        for w in wxApp.shownParamWidgets:
            if w[1].parameter.displayName == 'add' and self.first_call:
                self.add_old = w[1].paramChangeCtrl.GetValue()
                self.first_call = False

        self.plotSpecAdd()

        val_str = ""
        for v in self.vals:
            #val_str += str(v[0])+"-"+str(v[1])+","
            val_str += "%3.3f-%1.3e," % (v[0],v[1])

        new_params = list()
        new_params.append(reflex.RecipeParameter(
            recipe=self.recipe_name, displayName="add",
            value=self.add_old+val_str))
        return new_params


    def plotSpecAdd(self):
        x = [v[0] for v in self.vals]
        y = [v[1] for v in self.vals]
        self.p.sub_m.scatter(x, y, marker='+', s=300, linewidth=1.5, color='r')
        self.p.sub_d.scatter(x, y, marker='+', s=300, linewidth=1.5, color='r')


    def plotSpecDetail(self, ret=None):
        tooltip_d = ("Detail of the spectrum (frame category: RSPEC) with "
                     "detected lines (frame category: FLINE_PRE). Use the "
                     "slider below to shift the central wavelength of the "
                     "plotted region. Check `error` to overplot the error "
                     "spectrum. Middle-click to add lines. Added lines "
                     "appear as red crosses and will be added to the list of "
                     "detected lines when the recipe is re-run.")
        self.p.sub_d = self.p.sub(self.figure, self.gs[12:33,:], ret,
                                  tooltip=tooltip_d)
        self.rspec_d.step(self.p.sub_d, error=self.error_d)
        #self.rspec_smooth_d.step(self.p.sub_d)
        try:
            self.fline_pre_d.scatter(self.p.sub_d, marker='+', s=40,
                                     linewidth=1.5, color='g', zorder=10)
        except:
            pass
        self.p.sub_d.legend(prop={'size': 10})
        self.p.sub_d.callbacks.connect('xlim_changed', self.p.rectangle)
        self.p.sub_d.callbacks.connect('ylim_changed', self.p.rectangle)
        self.p.rectangle()
        self.plotSpecAdd()


    def plotSpecMain(self):
        tooltip_m = ("Spectrum (frame category: RSPEC) with detected lines "
                     "(frame category: FLINE_PRE). The light blue area "
                     "indicates the wavelength region plotted below.")
        try:
            self.p.sub_m = self.p.sub(self.figure, self.gs[:9,:],
                                      title="Detected lines (%i)"
                                      % self.fline_num, tooltip=tooltip_m)
        except:
            self.p.sub_m = self.p.sub(self.figure, self.gs[:9,:],
                                      title="Detected lines (0)",
                                      tooltip=tooltip_m)
        self.rspec_m.step(self.p.sub_m)
        #self.rspec_smooth_m.step(self.p.sub_m)
        try:
            self.fline_pre_m.scatter(self.p.sub_m, marker='+', s=40,
                                     linewidth=1.5, color='g', zorder=10)
        except:
            pass
        self.ylim_d = (self.p.sub_m.get_ylim()[0],
                       self.p.sub_m.get_ylim()[1])


    def plotWidgets(self):
        widgets = list()

        # Clickable plot
        self.first_call = True
        self.click = InteractiveClickableSubplot(self.p.sub_d, self.addLine)
        widgets.append(self.click)

        # Check buttons
        self.toggle = InteractiveCheckButtons(
            self.p.sub_b, self.togglePlotElem, ('error',), (0,))
        widgets.append(self.toggle)

        # Slider for detail
        self.det = Widget().slider(
            self.p.sub_s, self.refreshDetail, self.wavel_min, self.wavel_max,
            self.wavel_mean, "Wavel.", self.p.sub_m.get_xticks())
        widgets.append(self.det)

        return widgets


    def refreshDetail(self, wavel_sel):
        self.wavel_sel = wavel_sel
        self.rspec_d.select("WAVEL", (self.wavel_sel-10, self.wavel_sel+10))
        self.rspec_smooth_d.select("WAVEL",
                                   (self.wavel_sel-10, self.wavel_sel+10))
        self.fline_pre_d.select("WAVEL",
                                (self.wavel_sel-10, self.wavel_sel+10))
        self.p.sub_d.cla()
        self.plotSpecDetail(ret=self.p.sub_d)

    def togglePlotElem(self, sel):
        if sel == 'error':
            self.error_d = not(self.error_d)
        self.p.sub_d.cla()
        self.plotSpecDetail(ret=self.p.sub_d)


if __name__ == '__main__':
    import reflex_interactive_app
    import sys
    from espda_plot_common import *
    interactive_app = reflex_interactive_app.PipelineInteractiveApp(
        enable_init_sop=True)
    interactive_app.parse_args()
    if not import_success:
        interactive_app.setEnableGUI(False)
    if interactive_app.isGUIEnabled():
        dataPlotManager = DataPlotterManager()
        interactive_app.setPlotManager(dataPlotManager)
        from reflex_interactive_gui import ReflexInteractiveWxApp
        wxApp = ReflexInteractiveWxApp(interactive_app,
                                       interactive_app.dataPlotManager)
        add_start = list()
        for paramWidget in wxApp.shownParamWidgets:
            if paramWidget[1].parameter.displayName == "add":
                add_start = paramWidget[1].paramChangeCtrl.GetValue()
        wxApp.MainLoop()
    else:
        interactive_app.set_continue_mode()
    interactive_app.print_outputs()
    sys.exit()
