from __future__ import absolute_import
from __future__ import print_function
from six.moves import zip
try:
    import numpy as np
    import reflex
    from pipeline_product import PipelineProduct
    import pipeline_display
    import re
    import matplotlib.gridspec as gridspec
    import_success = True

except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")

#from copy import deepcopy as dc
from datetime import datetime

class DataPlotterManager:

    recipe_name = "espda_fit_qsocont"

    """ General methods """

    def addSubplots(self, figure):
        self.gs = gridspec.GridSpec(42, 6)
        self.figure = figure
        self.p = Plot()
        self.p.sub_b = self.p.sub(self.figure, self.gs[34:39,0:1], xlabel=None,
                                  ylabel=None)
        self.p.sub_s = self.p.sub(self.figure, self.gs[40:,:], xlabel=None,
                                  ylabel=None)


    def plotProductsGraphics(self):
        self.norm_d = False
        self.error_d = False
        self.fit_d = False
        self.plotSpec()


    def plotSpec(self):
        self.plotSpecMain()
        self.plotSpecDetail()


    def readFitsData(self, fitsFiles):
        frames = dict()
        self.names = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            frames[frame.category] = PipelineProduct(frame)
            self.names[frame.category] = frame.name

        wavel = frames["RSPEC_CONT"].readTableColumn(1, "WAVEL")
        self.wavel_min = np.min(wavel)
        self.wavel_max = np.max(wavel)
        self.wavel_mean = np.mean(wavel)

        self.rspec_cont_m = Data(frames["RSPEC_CONT"])
        #self.rspec_rem_m = Data(frames["RSPEC_REM"])
        self.fline_rem_m = Data(frames["FLINE_REM"], y="PEAK")
        self.spline_cont_m = Data(frames["SPLINE_CONT"])

        #self.rspec_cont_d = dc(self.rspec_cont_m)
        #self.rspec_rem_d = dc(self.rspec_rem_m)
        #self.fline_rem_d = dc(self.fline_rem_m)
        #self.spline_cont_d = dc(self.spline_cont_m)
        #self.spline_cont_new_d = dc(self.spline_cont_m)
        self.rspec_cont_d = Data(frames["RSPEC_CONT"])
        #self.rspec_rem_d = Data(frames["RSPEC_REM"])
        self.fline_rem_d = Data(frames["FLINE_REM"], y="PEAK")
        self.spline_cont_d = Data(frames["SPLINE_CONT"])
        self.spline_cont_new_d = Data(frames["SPLINE_CONT"])
        self.selectRange(self.wavel_mean)

    def selectRange(self, wavel):
        self.rspec_cont_d.select("WAVEL", (wavel-10, wavel+10))
        #self.rspec_rem_d.select("WAVEL", (wavel-10, wavel+10))
        self.fline_rem_d.select("WAVEL", (wavel-10, wavel+10))
        self.spline_cont_d.select("WAVEL", (wavel-10, wavel+10))
        self.spline_cont_new_d.select("WAVEL", (wavel-10, wavel+10))


    def setInteractiveParameters(self):
        paramList = list()
        #paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
        #    displayName = "hwidth",          group = "Line parameters",
        #    description = "Halfwidth for line grouping [nm]: [2.5e-2, 1.0e1]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "par-range",       group = "Line parameters",
            description = "Rules to set the range of line parameters\n" \
            "Sample syntax for par range: '1e-4,10.0,17.0,2.0,50.0,0.0,0.0', " \
            "meaning:\n" \
            "- redshift half-range: 1.0e-4\n" \
            "- minimum/maximum logarithimic column density: 10.0/17.0\n" \
            "- minimum/maximum Doppler broadening: 2.0/50.0\n" \
            "- minimum/maximum turbulence broadening: 0.0/0.0."))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "iter-max-temp",   group = "Line parameters",
            description = "Maximum number of iterations in the blue part: " \
            "[-1, 1e3]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "fit-flag",        group = "Fitting",
            description = "Flag to fit the lines"))
        #paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
        #    displayName = "peak-thres",      group = "Fitting",
        #    description = "Threshold to cut line peaks during during " \
        #    "normalization (as a fraction of the continuum level): [1.0e-2, " \
        #    "1.0]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "vel-sampl-blue",  group = "Fitting",
            description = "Velocity sampling to fit a spline to the blue part "\
            "[km/s]: [1.0e2, 1.0e4]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "vel-sampl-red",   group = "Fitting",
            description = "Velocity sampling to fit a spline to the red part " \
            "[km/s]: [1.0e2, 1.0e4]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "sg-width-pieces",   group = "Fitting",
            description = "Width of the Savitzky-Golay filter used to smooth "
            "the continuum in pieces [px]: [1e2, 1e4]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "sg-width-final",   group = "Fitting",
            description = "Width of the Savitzky-Golay filter used to smooth "
            "the final continuum [px]: [1e2, 1e4]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "model-resol",     group = "Fitting",
            description = "Resolution of the instrument model: [1.0e3, 1.0e6]"))
        paramList.append(reflex.RecipeParameter(recipe = self.recipe_name,
            displayName = "spline-knots",    group = "Line parameters",
            description = "Manually-entered spline knots"))
        return paramList


    def setWindowTitle(self):
        title = 'ESPRESSO Data Analysis Software: Continuum estimation'
        return title


    """ Specific methods """

    def corrSpline(self, point):
        wavel_sel, flux_sel, flux_denorm \
            = self.spline_cont_new_d.get_xy(point.xdata, point.ydata)
        displ = np.absolute(self.spline_cont_new_d.x-point.xdata)
        if np.min(displ) < 0.2:
            x_sel = np.argmin(displ)
            self.spline_cont_new_d.y[x_sel] = flux_denorm
        else:
            self.spline_cont_new_d.x = np.append(self.spline_cont_new_d.x,
                                                 point.xdata)
            self.spline_cont_new_d.y = np.append(self.spline_cont_new_d.y,
                                                 flux_denorm)
        try:
            self.spline_cont_new_d.select(None, (self.wavel_sel-10,
                                                 self.wavel_sel+10))
        except:
            self.spline_cont_new_d.select(None, (self.wavel_mean-10,
                                                 self.wavel_mean+10))

        self.p.sub_d.cla()
        self.plotSpecDetail(ret=self.p.sub_d)
        value = ""
        for w, f in zip(self.spline_cont_new_d.x, self.spline_cont_new_d.y):
            value += "%f:%e," % (w, f)
        path = re.sub("SPLINE_CONT.fits", "", self.names["SPLINE_CONT"])
        time = datetime.now().time().isoformat()
        name = path+"spline_knots_"+time+".txt"
        self.edit_file = open(name, "w")
        self.edit_file.write(value)
        self.edit_file.close()
        new_params = list()
        new_params.append(reflex.RecipeParameter(
            recipe=self.recipe_name, displayName="spline-knots", value=name))
        return new_params


    def plotSpecDetail(self, xlim=None, ylim=None, ret=None):
        tooltip_d = ("Detail of the continuum-fitted spectrum (frame category: "
                     "RSPEC_CONT) with removed lines (frame category: "
                     "FLINE_REM). The red dots are the spline knots used for "
                     "fitting. Use the slider below to shift the central "
                     "wavelength of the plotted region. Check `norm` to plot "
                     "the normalized spectrum. Check `error` to overplot the "
                     "error spectrum. Check `fit/resid.` to overplot the "
                     "spectrum of fitted lines and fitting residuals. "
                     "Middle-click at the wavelength of an existing nod to "
                     "adjust the knot position along the yaxis. Middle-click "
                     "between knots to add additional knots.")
        self.p.sub_d = self.p.sub(self.figure, self.gs[12:33,:], ret,
                                  tooltip=tooltip_d)
        if xlim != None:
            self.p.sub_d.set_xlim(xlim)
        if ylim != None:
            self.p.sub_d.set_xlim(ylim)
        self.rspec_cont_d.step(self.p.sub_d, xlim=xlim, error=self.error_d,
                               cont=True, fit=self.fit_d, resid=self.fit_d,
                               norm=self.norm_d)
        self.fline_rem_d.scatter(self.p.sub_d, norm=self.norm_d, marker='+',
                                 s=40, linewidth=1.5, color='g', zorder=10)
        #self.spline_cont_d.scatter(self.p.sub_d, norm=self.norm_d,
        #                           s=40, linewidth=1.5, color='g', zorder=10)
        self.spline_cont_new_d.scatter(self.p.sub_d, norm=self.norm_d,
                                       s=40, linewidth=1.5, color='r', zorder=9)
        hl = self.p.sub_d.get_legend_handles_labels()
        self.p.sub_d.legend(hl[0][:-1], hl[1][:-1],
                            prop={'size': 10}, loc=3)#,
                            #bbox_to_anchor=(1., 0.))
        if self.norm_d:
            ylim=self.p.sub_m.get_ylim()
        else:
            ylim=None
        self.p.sub_d.callbacks.connect('xlim_changed',
                                       lambda e: self.p.rectangle(e, ylim))
        self.p.sub_d.callbacks.connect('ylim_changed',
                                       lambda e: self.p.rectangle(e, ylim))
        self.p.rectangle(ylim=ylim)


    def plotSpecMain(self):
        tooltip_m = ("Continuum-fitted spectrum (frame category: RSPEC_CONT). "
                     "The light blue area indicates the wavelength region "
                     "plotted below.")
        self.p.sub_m = self.p.sub(self.figure, self.gs[:9,:],
                                  title="Continuum", tooltip=tooltip_m)
        self.rspec_cont_m.step(self.p.sub_m, cont=True)
        """
        self.fline_rem_m.scatter(self.p.sub_m, marker='+', s=40, linewidth=1.5,
                                 color='g', zorder=10)
        self.spline_cont_m.scatter(self.p.sub_m, s=40, linewidth=1.5,
                                   color='g', zorder=10)
        """
        self.ylim_d = (self.p.sub_m.get_ylim()[0],
                       self.p.sub_m.get_ylim()[1])

    def plotWidgets(self):
        widgets = list()

        # Clickable plot
        self.first_call = True
        self.click = InteractiveClickableSubplot(self.p.sub_d, self.corrSpline)
        widgets.append(self.click)

        # Check buttons
        self.toggle = InteractiveCheckButtons(
            self.p.sub_b, self.togglePlotElem, ('norm.', 'error', 'fit/resid.'),
            (0, 0, 0))
        widgets.append(self.toggle)

        # Slider for detail
        self.det = Widget().slider(
            self.p.sub_s, self.refreshDetail, self.wavel_min, self.wavel_max,
            self.wavel_mean, "Wavel.", self.p.sub_m.get_xticks())
        widgets.append(self.det)

        return widgets


    def refreshDetail(self, wavel_sel):
        self.wavel_sel = wavel_sel
        self.selectRange(self.wavel_sel)
        self.p.sub_d.cla()
        self.plotSpecDetail(ret=self.p.sub_d)


    def togglePlotElem(self, sel):
        if sel == 'norm.':
            self.norm_d = not(self.norm_d)
        elif sel == 'error':
            self.error_d = not(self.error_d)
        elif sel == 'fit/resid.':
            self.fit_d = not(self.fit_d)
        xlim = self.p.sub_d.get_xlim()
        ylim = self.p.sub_d.get_ylim()
        self.p.sub_d.cla()
        self.plotSpecDetail(xlim=xlim, ylim=ylim, ret=self.p.sub_d)


if __name__ == '__main__':
    import reflex_interactive_app
    import sys
    import wx
    from espda_plot_common import *
    interactive_app = reflex_interactive_app.PipelineInteractiveApp(
        enable_init_sop=True)
    interactive_app.parse_args()
    if not import_success:
        interactive_app.setEnableGUI(False)
    if interactive_app.isGUIEnabled():
        dataPlotManager = DataPlotterManager()
        interactive_app.setPlotManager(dataPlotManager)
        from reflex_interactive_gui import ReflexInteractiveWxApp
        wxApp = ReflexInteractiveWxApp(interactive_app,
                                       interactive_app.dataPlotManager)
        wxApp.MainLoop()
    else:
        interactive_app.set_continue_mode()
    interactive_app.print_outputs()
    sys.exit()
