from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from six.moves import range
from six.moves import zip
try:
    import numpy as np
    import reflex
    from pipeline_product import PipelineProduct
    import pipeline_display
    import re
    import matplotlib.gridspec as gridspec
    import_success = True
    import wx
    import wx.lib.mixins.listctrl as listmix
    from wx.lib.mixins.listctrl import CheckListCtrlMixin,ListCtrlAutoWidthMixin
    from datetime import datetime
    import sys

except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")


#from copy import deepcopy as dc
from matplotlib.ticker import MaxNLocator

class CheckListCtrl(wx.ListCtrl, CheckListCtrlMixin, ListCtrlAutoWidthMixin):
    def __init__(self, parent):
        wx.ListCtrl.__init__(self, parent, -1,
                             style=wx.LC_REPORT|wx.SUNKEN_BORDER)
        CheckListCtrlMixin.__init__(self)
        ListCtrlAutoWidthMixin.__init__(self)

class DataPlotterManager:

    recipe_name = "espda_iden_syst"

    """ General methods """

    def addSubplots(self, figure):
        self.gs = gridspec.GridSpec(42, 6)
        self.figure = figure
        self.p = Plot()

    def plotProductsGraphics(self):
        self.error_s = False
        self.fit_s = False
        self.vals = []
        self.sub_label = -1
        self.plotSpec()


    def plotSpec(self):
        self.plotSpecWhole()
        self.plotSpecSyst()

    def readFitsData(self, fitsFiles):
        frames = dict()
        for frame in fitsFiles:
            if frame == '':
                continue
            frames[frame.category] = PipelineProduct(frame)

        self.rspec_cont = Data(frames["RSPEC_CONT"])
        try:
            self.fline_iden = Data(frames["FLINE_IDEN"], y="PEAK")
        except:
            self.fline_iden = None
        self.qso_ion = Data(frames["QSO_ION"], y="OSCS")
        self.ion_wavel = {i.strip(): w \
                          for (i, w) in zip(self.qso_ion.l, self.qso_ion.x)}
        if self.fline_iden is not None:
            #self.fline_sel = dc(self.fline_iden)
            self.fline_sel = Data(frames["FLINE_IDEN"], y="PEAK")
            self.syst_num = len(set(self.fline_iden.s))
            self.systid = sorted(list(set(self.fline_iden.s)))
        else:
            self.fline_sel = None
            self.syst_num = 0
            self.systid = []
        if self.systid != []:
            self.systid_sel = str(self.systid[0])
        else:
            self.systid_sel = None

    def setInteractiveParameters(self):
        paramList = list()
        """
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName="grp-width",  group="Line grouping", description =
            "Width for redshift grouping [nm]: [1.0e-2, 1.0]"))
        """
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName="mult-dz",  group="Line grouping", description =
            "Minimum redshift difference between two candidate components of a "
            "doublet for the identification to be accepted : "
            "[1.0e-5, 1.0e-3]"))
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName="add-dv",  group="Line grouping", description =
            "Minimum velocity difference between velocity components of the "
            "same ionic species (e.g. CIV) two decide they belong to different "
            "absorption systems.: [1.0e-5, 1.0]"))
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName="grp-num",    group="Line grouping", description =
            "Minimum number of lines to define a group: [1, 10]"))
        #paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
        #    displayName="peak-thres",  group="Line grouping", description =
        #    "Peak threshold for starting the identification (normalized to "
        #    "the continuum): [0.0, 1.0]"))
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName="syst-dv", group="Line grouping", description =
            "Minimum velocity difference between different doublets/multiplets "
            "(e.g. SiIV and CIV, or MgII and FeII) to group them in the same "
            "system: [0.0, 1000.0]"))
        paramList.append(reflex.RecipeParameter(recipe=self.recipe_name,
            displayName="vis-rej", group="Visual selection", description =
            "List of visually selected systems"))
        return paramList


    """ Specific methods """

    def addLine(self, point):

        cond = np.where([point.inaxes.get_yaxis()==s.axes.get_yaxis() \
                         for s in self.p.sub_s])[0][0]
        val = self.rspec_cont.get_xy(point.xdata, point.ydata, self.v_s[cond])

        if val in self.vals:
            self.vals.remove(val)
        else:
            self.vals.append(val)

        """
        for w in wxApp.shownParamWidgets:
            if w[1].parameter.displayName == 'add' and self.first_call:
                self.add_old = w[1].paramChangeCtrl.GetValue()
                self.first_call = False
        """
        self.plotSpecAdd()
        """
        val_str = ""
        for v in self.vals:
            val_str += str(v[0])+"-"+str(v[1])+","
        """
        new_params = list()
        """
        new_params.append(reflex.RecipeParameter(
            recipe=self.recipe_name, displayName="add",
            value=self.add_old+val_str))
        """
        return new_params

    def plotSpecAdd(self):
        x = [v[0] for v in self.vals]
        y = [v[1] for v in self.vals]
        self.p.sub_w.scatter(x, y, marker='+', s=300, linewidth=1.5, color='r')
        for s, c in zip(self.p.sub_s, self.cen_s):
            s.scatter((1-c/x)*299792.458, y, marker='+', s=300,
                      linewidth=1.5, color='r')

    def plotSpecSyst(self, refresh_lineid=True, xlim=(-200,200),
                     ylim=(-0.5,1.5)):
        self.p.sub_w.legend(prop={'size': 10}, loc=3)
        self.sub_label += 1
        try:
            self.fline_sel_a.remove()
        except:
            pass
        if self.fline_sel is not None:
            self.fline_sel.select("SYSTID", self.systid_sel)
            self.fline_sel_a = self.fline_sel.scatter(
                self.p.sub_w, norm=True, marker='+', s=40, linewidth=1.5,
                color='g', zorder=10, label="selected")
            if refresh_lineid:
                self.lineid_sel = [l.strip() for l in np.unique(
                    self.fline_sel.l[self.fline_sel.sel])]
            self.systz_sel = np.mean(self.fline_sel.z[self.fline_sel.sel])
            self.wavel_sel = np.array([self.ion_wavel[i.strip()] \
                                       for i in self.lineid_sel])
        else:
            self.lineid_sel = []
            self.systz_sel = None
            self.wavel_sel = []

        ns = len(self.wavel_sel)
        nr = min(ns, 4)
        nc = min((ns-1)//4+1, 2)

        if ns > 8:
            print("Plotting only 8 panels")

        tooltip_s = ("Detail of the selected absorption system, split into "
                     "subplots of identified transitions. Check `error` to "
                     "overplot the error spectrum. Check `fit/resid.` to "
                     "overplot the spectrum of fitted lines and fitting "
                     "residuals. Use the `Systems` window to select systems. "
                     "Unchecked entries in the `Systems` window will be "
                     "removed when the recipe is re-run. Check entries in the "
                     "`Transitions` window to display other transitions and "
                     "look for possible lines to add to the sistems. "
                     "Middle-click on subplots to add lines. Added lines will "
                     "inherit the transition from the subplot (as denoted by "
                     "its label).")
        self.p.sub_s = []
        self.cen_s = []
        self.rspec_cont_a = []
        self.v_s = []
        for i, (l, w) \
            in enumerate(zip(self.lineid_sel, self.wavel_sel)):
            te = 12+(i%nr)*24//nr
            be = 12+(i%nr+1)*24//nr
            le = i//nr*6//nc
            re = (i//nr+1)*6//nc
            s = self.p.sub(self.figure, self.gs[te:be,le:re],
                           xlabel=r"Velocity difference [km/s]",
                           ylabel=None, label=self.sub_label, tooltip=tooltip_s)

            if i != 0:
                s.get_shared_x_axes().join(s, self.p.sub_s[0])
                s.get_shared_y_axes().join(s, self.p.sub_s[0])
            if i != nr-1 and i != ns-1:
                s.set_xticklabels([])
                s.set_xlabel("")
            if i >= nr:
                s.set_yticklabels([])
                s.set_ylabel("")
            if xlim != None:
                s.set_xlim(xlim)
            if ylim != None:
                s.set_ylim(ylim)

            s.yaxis.set_major_locator(MaxNLocator(prune='both'))
            c = (1+self.systz_sel)*w
            self.cen_s.append(c)
            self.p.sub_s.append(s)
            a, = self.rspec_cont.step(
                s, xlim=xlim, error=self.error_s, fit=self.fit_s,
                resid=self.fit_s, norm=True, cen=c, label=l)
            self.rspec_cont_a.append(a)
            self.v_s.append(self.rspec_cont.v)
            if self.fline_iden is not None:
                self.fline_iden.scatter(s, norm=True, cen=c, marker='+', s=40,
                                        linewidth=1.5, color='lightgray',
                                        zorder=9)
                self.fline_sel.scatter(s, norm=True, cen=c, marker='+', s=40,
                                       linewidth=1.5, color='g', zorder=10)
            s.text(0.05, 0.5, l, ha='left', va='center',
                   transform=s.transAxes)

        for s in self.p.sub_s:
            s.callbacks.connect(
                'xlim_changed',
                lambda e: self.p.rectangles(e, self.p.sub_s[0].get_ylim(),
                                            self.cen_s))

        if self.p.sub_s != []:
            self.p.rectangles(ylim=self.p.sub_s[0].get_ylim(), cen=self.cen_s)
            self.p.sub_b = self.p.sub(self.figure, self.gs[37:,0:1],
                                      xlabel=None, ylabel=None)


    def plotSpecWhole(self):
        tooltip_w = ("Continuum-fitted spectrum (frame category: RSPEC_CONT) "
                     "with identified (frame category: FLINE_IDEN) and "
                     "non-identified lines (frame category: FLINE_REJ).")
        self.p.sub_w = self.p.sub(self.figure, self.gs[:9,:],
                                  title="Identified systems (%i)" \
                                  % self.syst_num, ylabel="Normalized flux",
                                  tooltip=tooltip_w)
        self.rspec_cont.step(self.p.sub_w, norm=True)
        if self.fline_iden is not None:
            self.fline_iden.scatter(self.p.sub_w, norm=True, marker='+', s=40,
                                    linewidth=1.5, color='lightgray', zorder=9)

    def plotWidgets(self):
        widgets = list()

        # Clickable plot
        self.first_call = True
        self.click = []
        for s in self.p.sub_s:
            click = InteractiveClickableSubplot(s, self.addLine)
            self.click.append(click)
            widgets.append(click)

        # Check buttons
        if self.p.sub_s != []:
            self.toggle = InteractiveCheckButtons(
                self.p.sub_b, self.togglePlotElem, ('error', 'fit/resid.'),
                (0, 0))
            widgets.append(self.toggle)

        return widgets

    def togglePlotElem(self, sel):
        if sel == 'error':
            self.error_s = not(self.error_s)
        elif sel == 'fit/resid.':
            self.fit_s = not(self.fit_s)
        xlim = self.p.sub_s[0].get_xlim()
        ylim = self.p.sub_s[0].get_ylim()
        """
        for s in self.p.sub_s:
            d = self.rspec_cont
            if self.error_s:
                self.error_p = s.plot(d.x[d.sel]*d.v[d.sel],
                                      d.e[d.sel]*d.n[d.sel],
                                      label=d.label+" "+"error")
            else:
                try:
                    del self.error_p
                except:
                    pass
            if self.fit_s:
                self.fit_p = s.plot(d.x[d.sel]*d.v[d.sel],
                                    d.f[d.sel]*d.n[d.sel],
                                    label=d.label+" "+"fit")
                self.resid_p = s.plot(d.x[d.sel]*d.v[d.sel],
                                      d.r[d.sel]*d.n[d.sel],
                                      label=d.label+" "+"resid.")
            else:
                try:
                    del self.fit_p
                    del self.resid_p
                except:
                    pass
        """
        for s in self.p.sub_s:
            s.cla()
        self.plotSpecSyst(xlim=xlim, ylim=ylim, refresh_lineid=False)

class IonPanelManager(wx.Frame):
    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, size = (454, 300),
            pos = (0, 400))
        self.dpm = dataPlotManager

        panel = wx.Panel(self, -1)
        vbox = wx.BoxSizer(wx.VERTICAL)

        hbox1 = wx.BoxSizer(wx.HORIZONTAL)
        top_panel = wx.Panel(panel, -1)
        self.cl = CheckListCtrl(top_panel)
        self.cl.InsertColumn(0, "", width = 23)
        self.cl.InsertColumn(1, "Line ID", width = 100)
        self.cl.InsertColumn(2, "Rest-frame", width = 100)
        self.cl.InsertColumn(3, "Redshifted", width = 100)
        self.cl.InsertColumn(4, "Osc. strength", width = 100)
        hbox1.Add(self.cl, 1, wx.EXPAND)
        top_panel.SetSizer(hbox1)
        vbox.Add(top_panel, 1, wx.EXPAND | wx.ALL, 5)
        panel.SetSizer(vbox)
        self.fill()
        self.cl.OnCheckItem = self.onCheckItem


    def fill(self):
        self.selected = []
        wavel_min = np.min(self.dpm.rspec_cont.x)
        wavel_max = np.max(self.dpm.rspec_cont.x)
        ion = self.dpm.qso_ion
        if self.dpm.systz_sel is not None:
            sel = np.where(np.logical_and(
                ion.x*(1+self.dpm.systz_sel)>wavel_min,
                ion.x*(1+self.dpm.systz_sel)<wavel_max))
        else:
            sel = []
        for l, x, y \
            in zip(ion.l[sel], ion.x[sel], ion.y[sel]):
            try:
                index = self.cl.InsertItem(sys.maxsize, "")
                self.cl.SetItem(index, 1, str(l))
                self.cl.SetItem(index, 2, str(x))
                self.cl.SetItem(index, 3, str(x*(1+self.dpm.systz_sel)))
                self.cl.SetItem(index, 4, str(y))
            except:  # Python 2
                index = self.cl.InsertStringItem(sys.maxsize, "")
                self.cl.SetStringItem(index, 1, str(l))
                self.cl.SetStringItem(index, 2, str(x))
                self.cl.SetStringItem(index, 3, str(x*(1+self.dpm.systz_sel)))
                self.cl.SetStringItem(index, 4, str(y))
            if l.strip() in self.dpm.lineid_sel:
                self.selected.append(index)
                self.cl.CheckItem(index)
        self.Show()


    def onCheckItem(self, index, flag):
        i = str(self.cl.GetItem(index, 1).GetText()).strip()
        if i in self.dpm.lineid_sel:
            self.dpm.lineid_sel.remove(i)
            self.selected.remove(index)
        else:
            self.dpm.lineid_sel.append(i)
            self.selected.append(index)
        if len(self.dpm.lineid_sel) > 8:
            self.cl.CheckItem(self.selected[0], check=False)
            self.selected = self.selected[-8:]
            self.dpm.lineid_sel = self.dpm.lineid_sel[-8:]
        for s in self.dpm.p.sub_s:
            wxApp.figure.delaxes(s)
        self.dpm.plotSpecSyst(refresh_lineid=False)
        wxApp.figureCanvas.draw()


class SystPanelManager(wx.Frame):
    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, size=(290, 300),
            pos=(0, 0))

        self.dpm = dataPlotManager
        self.ipm = ionPanelManager

        panel = wx.Panel(self, -1)
        vbox = wx.BoxSizer(wx.VERTICAL)

        hbox1 = wx.BoxSizer(wx.HORIZONTAL)
        top_panel = wx.Panel(panel, -1)
        self.cl = CheckListCtrl(top_panel)
        self.cl.InsertColumn(0, "", width=23)
        self.cl.InsertColumn(1, "ID",   width=50)
        self.cl.InsertColumn(2, "Redshift",    width=80)
        self.cl.InsertColumn(3, "Trans.", width=50)
        self.cl.InsertColumn(3, "Comp.",  width=50)
        hbox1.Add(self.cl, 1, wx.EXPAND)
        top_panel.SetSizer(hbox1)
        vbox.Add(top_panel, 1, wx.EXPAND | wx.ALL, 5)
        panel.SetSizer(vbox)

        for widget in wxApp.shownParamWidgets:
            if widget[1].parameter.displayName == "vis-rej":
                vis_rej = widget[1].paramChangeCtrl.GetValue()

        vis_value = vis_rej.split(",")

        if vis_rej == "":
            self.systid_rej = []
        else:
            self.systid_rej = vis_rej.split(",")
        for i in range(0, dataPlotManager.syst_num):
            sel = np.where(self.dpm.fline_iden.s == self.dpm.systid[i])[0]
            try:
                index = self.cl.InsertItem(sys.maxsize, "")
                self.cl.SetItem(index, 1, self.dpm.systid[i])
                self.cl.SetItem(
                    index, 2,
                    "{:4.6f}".format(np.mean(self.dpm.fline_iden.z[sel])))
                self.cl.SetItem(
                    index, 3, str(len(self.dpm.fline_iden.l[sel])))
                self.cl.SetItem(
                    index, 4, str(len(list(set(self.dpm.fline_iden.l[sel])))))
            except:  # Python 2
                index = self.cl.InsertStringItem(sys.maxsize, "")
                self.cl.SetStringItem(index, 1, self.dpm.systid[i])
                self.cl.SetStringItem(
                    index, 2,
                    "{:4.6f}".format(np.mean(self.dpm.fline_iden.z[sel])))
                self.cl.SetStringItem(
                    index, 3, str(len(self.dpm.fline_iden.l[sel])))
                self.cl.SetStringItem(
                    index, 4, str(len(list(set(self.dpm.fline_iden.l[sel])))))
            if self.dpm.systid[i] not in self.systid_rej:
                self.cl.CheckItem(index)
            if i is 0:
                self.cl.Select(index)
        self.Show()
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onSelItem)
        self.cl.OnCheckItem = self.onCheckItem


    def onNull(self, event, flag):
        pass

    def onSelItem(self, event):
        sel_index = self.cl.GetNextSelected(-1)
        self.dpm.systid_sel = str(self.cl.GetItem(sel_index, 1).GetText())
        for s in self.dpm.p.sub_s:
            wxApp.figure.delaxes(s)
        self.dpm.plotSpecSyst()
        self.ipm.cl.OnCheckItem = self.onNull
        self.ipm.cl.DeleteAllItems()
        self.ipm.fill()
        self.ipm.cl.OnCheckItem = self.ipm.onCheckItem

        wxApp.figureCanvas.draw()

    def onCheckItem(self, index, flag):
        if flag is False:
            self.systid_rej.append(self.dpm.systid[index])
        else:
            self.systid_rej.remove(self.dpm.systid[index])
        for widget in wxApp.shownParamWidgets:
            if widget[1].parameter.displayName == "vis-rej":
                old = widget[1].paramChangeCtrl.GetValue()
                widget[1].paramChangeCtrl.SetValue(','.join(self.systid_rej) \
                                                   + ",")


if __name__ == '__main__':
    import reflex_interactive_app
    import sys
    import wx
    from espda_plot_common import *
    interactive_app = reflex_interactive_app.PipelineInteractiveApp(
        enable_init_sop=True)
    interactive_app.parse_args()
    if not import_success:
        interactive_app.setEnableGUI(False)
    if interactive_app.isGUIEnabled():
        dataPlotManager = DataPlotterManager()
        interactive_app.setPlotManager(dataPlotManager)
        from reflex_interactive_gui import ReflexInteractiveWxApp
        wxApp = ReflexInteractiveWxApp(interactive_app,
                                       interactive_app.dataPlotManager)
        try:
            ionPanelManager = IonPanelManager(wxApp.frame, -1, "Transitions")
            systPanelManager = SystPanelManager(wxApp.frame, -1, "Systems")
        except:
            pass
        wxApp.MainLoop()
    else:
        interactive_app.set_continue_mode()
    interactive_app.print_outputs()
    sys.exit()
